/*-
 * Copyright (c) 2003 the grugq
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the grugq
 * 4. The name of the author may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>

static void
next_line(char *line, char **next)
{
	while (*line++ != '\n')
		;
	*next = line;
}

static size_t
strtoi(char *str, char **next)
{
	size_t	val = 0, accum = 0;

	while (*str) {
		val *= 16;

		switch (*str) {
		case '0': val += 0; break;
		case '1': val += 1; break;
		case '2': val += 2; break;
		case '3': val += 3; break;
		case '4': val += 4; break;
		case '5': val += 5; break;
		case '6': val += 6; break;
		case '7': val += 7; break;
		case '8': val += 8; break;
		case '9': val += 9; break;
		case 'a':
		case 'A': val += 10; break;
		case 'b':
		case 'B': val += 11; break;
		case 'c':
		case 'C': val += 12; break;
		case 'd':
		case 'D': val += 13; break;
		case 'e':
		case 'E': val += 14; break;
		case 'f':
		case 'F': val += 15; break;
		default:
			  goto out;
		}
		accum = val;
		str++;
	}
out:
	if (next)
		*next = str;
	return (accum);
}

/* lline looks like:
 * 08048000-08049000 .... */
static void *
extract(char *line, size_t *len)
{
	char	* np;
	size_t	  base, end;


	base = strtoi(line, &np);
	np++;

	end = strtoi(np, NULL);

	*len = end - base;

	return (void *) base;
}

int
ul_map_down(void)
{
	ssize_t	n;
	char	buf[512], * str;
	int	fd, i;


	if ((fd = open("/proc/self/maps", O_RDONLY)) == -1)
		return (-1);

	memset(buf, 0x00, sizeof(buf));
	if ((n = read(fd, buf, sizeof(buf))) < 0) {
		close(fd);
		return (-1);
	}
	close(fd);

	for (str = buf, i = 0; *str && i < 3; next_line(str, &str), i++) {
		void	* addr;
		size_t	  len;

		addr = extract(str, &len);

		if (munmap(addr, len))
			return (-1);
	}

	return (0);
}
